#include <iostream>
#include <atomic>
#include <thread>
#include <vector>
#include <chrono>
#include <random>
#include <future>

// Thread-safe counter class
class Counter {
private:
    std::atomic<int> value;
public:
    Counter() : value(0) {}
    int next() {
        return ++value;
    }
};

// Thread-safe sum class
class SumAccumulator {
private:
    std::atomic<long long> total;
public:
    SumAccumulator() : total(0) {}
    void add(int value) {
        total.fetch_add(value, std::memory_order_relaxed);
    }
    long long getTotal() const {
        return total.load(std::memory_order_relaxed);
    }
};

// Service function that reads the counter value, sleeps, and returns the value
int service(Counter& counter) {
    int value = counter.next();
    // Sleep for a random amount of time up to 1 second
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<> dis(1, 1000);
    std::this_thread::sleep_for(std::chrono::milliseconds(dis(gen)));
    return value;
}

int main() {
    Counter counter;
    SumAccumulator sumAccumulator;

    const int totalCalls = 1000; // 1 million times

    // Vector to hold future objects
    std::vector<std::future<void>> futures;

    auto startTime = std::chrono::high_resolution_clock::now();

    // Launching async tasks
    for (int i = 0; i < totalCalls; ++i) {
        futures.push_back(std::async(std::launch::async, [&counter, &sumAccumulator] {
            int value = service(counter);
            sumAccumulator.add(value);
        }));
    }

    // Wait for all tasks to complete
    for (auto& fut : futures) {
        fut.get();
    }

    auto endTime = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> duration = endTime - startTime;

    // Final total should be 500,000,500,000
    std::cout << "Final Total: " << sumAccumulator.getTotal() << std::endl;
    std::cout << "Execution Time: " << duration.count() << " seconds" << std::endl;

    return 0;
}
